package NAT::PTD;

#------------------------------------------------------------------#
# NATools - Package with parallel corpora tools                    #
# Copyright (C) 2008  Alberto Simões                               #
#                                                                  #
# This package is free software; you can redistribute it and/or    #
# modify it under the terms of the GNU Lesser General Public       #
# License as published by the Free Software Foundation; either     #
# version 2 of the License, or (at your option) any later version. #
#                                                                  #
# This library is distributed in the hope that it will be useful,  #
# but WITHOUT ANY WARRANTY; without even the implied warranty of   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU #
# Lesser General Public License for more details.                  #
#                                                                  #
# You should have received a copy of the GNU Lesser General Public #
# License along with this library; if not, write to the            #
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,     #
# Boston, MA 02111-1307, USA.                                      #
#------------------------------------------------------------------#

use base 'Exporter';
our @EXPORT = 'toentry';
use warnings;
use strict;


=head1 NAME

NAT::PTD - Module to handle PTD files in Dumper Format

=head1 SYNOPSIS

  use NAT::PTD;

	$ptd = new NAT::PTD( $ptd_file );

=head1 DESCRIPTION

PTD files in Perl Dumper format are simple hashes references. But they
use a specific structure, and this module provides a simple interface to
manipulate it.

=head2 C<new>

The C<new> constructor returns a new NAT::PTD object. This constructor
receives a PTD file in dumper format.

  $ptd = new NAT::PTD( $ptd_file );

=cut

sub new {
	my ($class, $dmp) = @_;
	return undef unless -f $dmp;
	
	my $self = do $dmp;
	$self = bless $self => $class; #amen
	$self->_calculate_size();
	return $self;
}

=head2 C<dump>

The C<dump> method is used to write the PTD in its own format, but
taking care to sort words lexicographically, and sorting translations by
their probability (starting with higher probabilities).

The format is Perl code, and thus, can be used independetly of this module.

   $ptd -> dump;

Note that the C<dump> method writes to the Standard Output stream.

=cut

sub dump {
	my $self = shift;

	print "\$a = {\n";
	$self->downtr(
		sub {
			my ($w,$c,%t) = @_;			
			printf "  '%s' => {\n", _protect_quotes($w);
			printf "      count => %d,\n", $c;
			printf "      trans => {\n";
			for my $t (sort { $t{$b} <=> $t{$a} } keys %t) {
				printf "          '%s' => %.6f,\n", _protect_quotes($t), $t{$t};
			}
			printf "      }\n";
			printf "  },\n";
		},
		sorted => 1,
	);

	print "}\n";
}

=head2 C<words>

The C<words> method returns an array (not a reference) to the list of
words of the dictionary: its domain.

   @words = $ptd -> words;

=cut

sub words {
	my $self = shift;
	return grep {!/^ /} keys %$self;
}

=head2 C<trans>

The C<trans> method receives a word, and returns the list of its possible
translations.

   @translations = $ptd -> trans( "dog" );

=cut

sub trans {
	my ($self, $word) = @_;
	return () unless exists $self->{$word};
	return keys %{$self->{$word}{trans}};
}

=head2 C<prob>

The C<prob> method receives a word and a translation, and returns the
probability of that word being translated that way.

   $probability = $ptd -> prob ( "cat", "gato");

=cut

sub prob {
	my ($self, $word, $trad) = @_;
	return 0 unless exists $self->{$word}{trans}{$trad};
	return $self->{$word}{trans}{$trad};
}

=head2 C<size>

Returns the total number of words from the source-corpus that originated
the PTD. Basically, the sum of the C<count> attribute for all words.

   $size = ptd -> size;

=cut

sub size {
	return $_[0]->{' size '}; # space is relevant
}

=head2 C<count>

The C<count> method receives a word and returns the occurrence count for
that word.

   $count = $ptd -> count ("cat");

=cut

sub count {
	if (exists $_[0]->{$_[1]}) {
		return $_[0]->{$_[1]}{count}
	} else {
		return 0;
	}
}

=head2 C<reprob>

This method recalculates all probabilities accordingly with the number
of translations available.

For instance, if you have

    home => casa => 25%
         => lar  => 25%

The resulting dictionary will have

   home => casa => 50%
        => lar  => 50%

Note that this methods B<replaces> the object.

=cut

sub reprob {
	my $self = shift;
	
	$self -> downtr (
		sub {
			my ($w, $c, %t) = @_;
			my $actual;
			$actual += $t{$_} for (keys %t);
			$t{$_} /= $actual for (keys %t);
			return toentry($w, $c, %t);
		},
		filter => 1
	);
	
	return $self;
}

=head2 C<intersect>

This method intersects the current object with the supplied PTD.
Note that this method B<replaces> the object values.

=cut

sub intersect {
	my ($self, $other) = @_;
	
	for my $w (keys %$self) {
		next if $w =~ /^ /;
		if (exists($other->{$w})) {
			$self->{$w}{count} = _min($self->count($w), $other->count($w));
			for my $t (keys %{$self->{$w}{trans}}) {
				if (exists($other->{$w}{trans}{$t})) {
					$self->{$w}{trans}{$t} = _min($self->trans($w,$t), $other->trans($w,$t))
				}
				else {
					delete $self->{$w}{trans}{$t}
				}
			}
		}
		else {
			delete $self->{$w}
		}
	}
	$self->_calculate_size();
}

=head2 C<add>

This method adds the current PTD with the supplied one.
Note that this method B<replaces> the object values.

=cut

sub add {
	my ($self, $other) = @_;
	
	my ($S1,$S2) = ($self->size,  $other->size);
	my @words = ($self->words, $other->words);
	{
		# this way %words get freed ASAP.
		my %words;
		@words{@words}=@words;
		@words = keys %words;		
	}

	for my $w (@words) {
		my ($c1, $c2) = ($self->count($w), $other->count($w));
		
		my %trans;
		for my $t ($self->trans($w),$other->trans($w)) {
			next if exists $trans{$t}; ## do not process the same word twice
			
			my ($p1, $p2) = ($self->prob($w,$t), $other->prob($w,$t));
			my ($w1, $w2) = ($c1*$S2, $c2*$S1);
			if ($w1 + $w2) {
				$trans{$t} = ($w1 * $p1 + $w2 * $p2)/($w1 + $w2);
			}
		}
		$self->{$w}{trans} = \%trans;
		$self->{$w}{count} = $c1 + $c2;
	}
	$self->_calculate_size();
	return $self;
}

=head2 C<downtr>

This method iterates over a dictionary and calls the function supplied
as argument. This function will receive, in each call, the word in the
source language, the number of occurrences, and the hash of
translations.

  $ptd -> downtr( sub { my ($w,$c,%t) = @_;
                        if ($w =~ /[^A-Za-z0-9]/) {
                            return undef;
                        } else {
                            return toentry($w,$c,%t);
                        }
                });

=cut

sub downtr {
	my ($self, $sub, %opt) = @_;

	my @keys = $opt{sorted} ? sort keys %$self : keys %$self;
	for my $word (@keys) {
		next if $word =~ /^ /;
		my $res = $sub->($word,
						 $self->{$word}{count},
						 %{$self->{$word}{trans}});
		if ($opt{filter}) {
			if (!defined($res)) {
				delete $self->{$word}
			} else {
				$self->{$word} = $res;
			}			
		}
	}
}

=head2 C<toentry>

This function is exported by default and creates a dictionary entry
given the word, word count, and hash of translations. Check C<downtr>
for an example.

=cut

sub toentry {
	## word (ignored), count, ref(%hash)
	if (ref($_[2]) eq "HASH") {
		return { count => $_[1], trans => $_[2] }		
	}
	else {
		my ($w, $c, %t) = @_;
		return { count => $c, trans => \%t }		
	}

}

=head1 SEE ALSO

NATools, perl(1)

=head1 AUTHOR

Alberto Manuel Brandão Simões, E<lt>ambs@cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2008 by Alberto Manuel Brandão Simões

=cut

sub _calculate_size {
	my $self = shift;
	my $total = 0;
	$self->downtr( sub { $total += $_[1] });
	$self->{" size "} = $total;           ## Private keys are kept with spaces.
}

sub _min { $_[0] < $_[1] ? $_[0] : $_[1] }

sub _protect_quotes {
	my $f = shift;
	for ($f) {
		s/\\/\\\\/g;
		s/'/\\'/g;
	}
	return $f;
}

"This isn't right.  This isn't even wrong.";
__END__
